from dataclasses import dataclass, field
from datetime import date

from dataclass_wizard import DumpMeta, YAMLWizard
from yaml.composer import ComposerError

from . import VERSION
from .utils import strip_html_tags


@dataclass
class Author:
    forname: str
    surname: str
    orcidurl: str = ""
    presentation: str = ""
    display: bool = None
    institution: str = ""
    foaf: str = ""
    isni: str = ""
    orcid: str = ""
    viaf: str = ""
    wikidata: str = ""
    affiliations: str = ""
    email: str = ""
    biography: str = ""


@dataclass
class Book(YAMLWizard):
    version: int
    title: str
    lang: str
    date: date
    rights: str
    url: str
    collective: bool
    coverurl: str
    abstract_fr: str
    abstract_en: str
    keyword_fr: str
    keyword_en: str
    isbnprint: str
    isbnepub: str
    isbnpdf: str
    isbnnum: str
    publisher: str
    place: str

    # (Re)Filled in a second step.
    authors: list
    toc: list
    chapters: list = None
    chapters_and_parts: list = None
    has_search: bool = False
    title_h: str = None
    title_f: str = None
    subtitle: str = None
    subtitle_f: str = None
    nbpages: str = None
    date_french: str = None
    zoteroDescription: str = "Accéder à cette bibliographie sur Zotero"
    pressoir_version: str = VERSION
    url_publisher: str = ""
    productor: str = ""
    productor_num: str = ""
    diffusor: str = ""
    pdfurl: str = ""
    epuburl: str = ""
    luluurl: str = ""

    year: str = ""
    month: str = ""
    day: str = ""

    @property
    def has_parts(self):
        return len(self.chapters) != len(self.chapters_and_parts)

    def __post_init__(self):
        self.title_h = self.title_h or self.title
        self.title_f = self.title_f or strip_html_tags(self.title)
        self.subtitle_f = self.subtitle_f or strip_html_tags(self.subtitle)

        dataclass_authors = []
        for author in self.authors:
            dataclass_authors.append(Author(**author))
        self.authors = dataclass_authors

        self.year, self.month, self.day = self.date.isoformat().split("-")


@dataclass
class Part:
    title: str
    chapters: list = None


@dataclass
class Chapter(YAMLWizard):
    title: str
    abstract_fr: str = ""
    abstract_en: str = ""
    keyword_fr: str = ""
    keyword_en: str = ""

    # (Re)Filled in a second step.
    authors: list = field(default_factory=list)
    translators: list = field(default_factory=list)

    id: str = None
    title_h: str = None
    title_f: str = None
    subtitle: str = None
    subtitle_f: str = None
    part: Part = None
    blockcitation: bool = None
    pdf: bool = True
    is_not_search: bool = True
    url_traduction: str = ""
    url: str = ""
    url_relative: str = ""
    zoteroCollection: str = ""
    lang: str = ""
    search_data: dict = field(default_factory=dict)

    def __post_init__(self):
        self.title_h = self.title_h or self.title
        self.title_f = self.title_f or strip_html_tags(self.title)
        self.subtitle_f = self.subtitle_f or strip_html_tags(self.subtitle)

        dataclass_authors = []
        for author in self.authors:
            dataclass_authors.append(Author(**author))
        self.authors = dataclass_authors

        dataclass_translators = []
        for translator in self.translators:
            dataclass_translators.append(Author(**translator))
        self.translators = dataclass_translators


def configure_book(yaml_path):
    # Preserves abstract_fr key for instance (vs. abstract-fr) when converting to_yaml()
    DumpMeta(key_transform="SNAKE").bind_to(Book)
    DumpMeta(key_transform="SNAKE").bind_to(Chapter)

    try:
        book = Book.from_yaml_file(yaml_path)
    except ComposerError:
        book = Book.from_yaml(yaml_path.read_text().split("---")[1])
    repository_path = yaml_path.parent.parent
    book.chapters_and_parts = configure_chapters_and_parts(book, repository_path)
    only_chapters = []
    for chapter_or_part in book.chapters_and_parts:
        if isinstance(chapter_or_part, Chapter):
            only_chapters.append(chapter_or_part)
        elif isinstance(chapter_or_part, Part):
            for chapter in chapter_or_part.chapters:
                only_chapters.append(chapter)
    book.chapters = only_chapters
    book.has_search = "recherche" in [chap.id for chap in book.chapters]
    return book


def configure_chapters_and_parts(book, repository_path):
    dataclass_chapters = []
    for chapter in book.toc:
        if "id" in chapter:
            dataclass_chapters.append(
                configure_chapter(book, chapter["id"], repository_path)
            )
        elif "parttitle" in chapter:
            part = Part(title=chapter["parttitle"])
            dataclass_chapters.append(part)
            part_chapters = []
            for chap in chapter["content"]:
                part_chapters.append(
                    configure_chapter(book, chap["id"], repository_path)
                )
            part.chapters = part_chapters
    return dataclass_chapters


def configure_chapter(book, chapter_id, repository_path):
    try:
        chapter = Chapter.from_yaml_file(
            repository_path / chapter_id / f"{chapter_id}.yaml"
        )
    except ComposerError:
        chapter = Chapter.from_yaml(
            (repository_path / chapter_id / f"{chapter_id}.yaml")
            .read_text()
            .split("---")[1]
        )

    chapter.id = chapter_id
    chapter.url = f"{book.url}{chapter_id}.html"
    chapter.url_relative = f"{chapter_id}.html"
    chapter.is_not_search = "recherche" != chapter_id
    return chapter



import locale
from collections import defaultdict

import unidecode
from jinja2 import Environment as Env
from jinja2 import FileSystemLoader
from jinja2.exceptions import FilterArgumentError
from jinja2.filters import ignore_case
from selectolax.parser import HTMLParser

from . import ROOT_DIR
from .utils import generate_md5

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

# This is the jinja2 configuration to locate templates.
environment = Env(loader=FileSystemLoader(str(ROOT_DIR / "templates")))
# Useful for dates rendering within Jinja2.
locale.setlocale(locale.LC_ALL, "fr_FR.UTF-8")


def do_dictsortunicode(value, case_sensitive=False, by="key", reverse=False):
    """Same as built-in dictsort but order accentuated words the right way.

    Inspired by
    https://github.com/pallets/jinja/blob/27c65757b26bb5012df1a5ccab1340cd7d52f139/
    src/jinja2/filters.py#L257
    """
    if by == "key":
        pos = 0
    elif by == "value":
        pos = 1
    else:
        raise FilterArgumentError('You can only sort by either "key" or "value"')

    def sort_func(item):
        value = item[pos]

        if not case_sensitive:
            value = ignore_case(value)

        # This is where we remove accents from the string.
        value = unidecode.unidecode(value)

        # And where we strip the markdown emphasis.
        value = value.replace("*", "")

        return value

    return sorted(value.items(), key=sort_func, reverse=reverse)


environment.filters["dictsortunicode"] = do_dictsortunicode


def do_markdownem(value):
    if "*" in value:
        pre, text = value.split("*", 1)
        if "*" in text:
            term, post = text.split("*", 1)
            value = f"{pre}<em>{term}</em>{post}"
    return value


environment.filters["markdownem"] = do_markdownem


def collect_index_and_update_attrs(chapter_id, html, kind, ids, refs):
    for item in html.css(f"h4.{kind},span.{kind}"):
        idsp = item.attributes.get("data-idsp")
        glossaire_id = item.attributes.get("data-idglossaire")
        if idsp is None:
            idsp = item.text()
        idsp_id = f"{kind}-{generate_md5(idsp)}"
        # In place modification of the item by selectolax (class addition).
        item.attrs["class"] = f"{item.attrs['class']} {idsp_id}"
        if chapter_id in refs[idsp]:
            refs[idsp][chapter_id] += 1
            # Ensure we only add the same id once per page.
            continue
        refs[idsp][chapter_id] = 1
        ids[kind][idsp] = {
            "idsp_id": idsp_id,
            "glossaire_id": glossaire_id,
        }
        # In place modification of the item by selectolax (id addition).
        item.attrs["id"] = idsp_id


def collect_indexes_and_add_ids(target, book, kinds):
    ids = defaultdict(dict)
    refs = defaultdict(dict)
    for chapter in book.chapters:
        chapter_id = chapter.id
        html_file = target / f"{chapter_id}.html"
        if not html_file.exists():
            break
        html_content = html_file.read_text()
        parsed_html = HTMLParser(html_content)
        for kind, title in kinds.items():
            collect_index_and_update_attrs(chapter_id, parsed_html, kind, ids, refs)
        html_file.write_text(parsed_html.html)
    return ids, refs


def generate_index(ids, refs, kinds):
    template = environment.get_template("index.html")
    return template.render(ids=ids, refs=refs, kinds=kinds)


def write_index(output, content):
    index_html = output.read_text()
    index_html = index_html.replace("%INDEX%", content)
    output.write_text(index_html)


def generate_indexes(repository_path, target_path, book):
    """Generate the index page and related ids."""
    print(f"Generating indexes for {repository_path}")
    book_settings = tomllib.loads(
        (repository_path / "pressoir" / "book.toml").read_text()
    )
    kinds = dict(
        zip(book_settings["indexes"]["ids"], book_settings["indexes"]["names"])
    )
    ids, refs = collect_indexes_and_add_ids(target_path, book, kinds)
    index_content = generate_index(ids, refs, kinds)
    index_file = target_path / "index-np.html"
    if index_file.exists():
        write_index(index_file, index_content)



#!/usr/bin/env python3
import html
import tempfile
from collections import namedtuple

import pypandoc
import yaml
from progressist import ProgressBar

from .utils import get_template_path, strip_html_comments

AdditionalContent = namedtuple("AdditionalContent", "id, title, type, yaml, md, html")
SEPARATOR = "## "


def extract_additional_data(repository_path, chapter):
    additional_path = repository_path / "textes" / chapter.id / "additionnels.md"
    if not additional_path.exists():
        return False
    additional_data = additional_path.read_text()
    if not additional_data.strip() or SEPARATOR not in additional_data:
        return False
    return additional_data


def extract_additional_content(repository_path, chapter, additional_item):
    try:
        additional_id, yaml_content, md_content = additional_item.split("---", 2)
    except ValueError as error:
        print(f"    /!\\ Error in {chapter.id}/additionels.md (malformed file)")
        print(additional_item)
        print(error)
        return False

    additional_id = additional_id.strip()
    if additional_id.startswith(SEPARATOR):
        additional_id = strip_html_comments(additional_id[len(SEPARATOR) :]).strip()

    try:
        yaml_content_loaded = yaml.load(yaml_content, Loader=yaml.Loader)
    except yaml.scanner.ScannerError as error:
        print(f"    /!\\ Error in {chapter.id}/additionels.md (malformed file)")
        print(error)
        return False

    additional_title = yaml_content_loaded.get("title")
    additional_type = yaml_content_loaded.get("type")

    if additional_type is None:
        return False

    if additional_type.startswith("texte"):
        additional_type = "texte"
    elif additional_type in ["entretien", "situation", "tableau"]:
        additional_type = "article"
    template_path = get_template_path(repository_path, f"{additional_type}.html")
    with tempfile.NamedTemporaryFile() as metadata_file:
        yaml_content += "suppress-bibliography: true\n"
        metadata_file.write(yaml_content.encode("utf-8"))
        metadata_file.read()  # Required to be readable by Pandoc.
        bib_file = repository_path / "textes" / chapter.id / f"{chapter.id}.bib"
        html_content = pypandoc.convert_text(
            md_content,
            "html",
            format="md",
            extra_args=[
                "--ascii",
                "--citeproc",
                f"--bibliography={bib_file}",
                f"--template={template_path}",
                f"--metadata-file={metadata_file.name}",
                f"--variable=additional_id:{additional_id}",
            ],
        )
    return AdditionalContent(
        **{
            "id": additional_id,
            "title": additional_title,
            "type": additional_type,
            "yaml": yaml_content,
            "md": md_content,
            "html": html_content,
        }
    )


def extract_additional_contents(repository_path, chapter):
    additional_contents = []
    additional_data = extract_additional_data(repository_path, chapter)
    if not additional_data:
        return []
    additional_parts = additional_data.split(f"\n{SEPARATOR}")
    bar = ProgressBar(
        total=len(additional_parts), prefix="    Loading additional contents:"
    )
    for additional_item in bar.iter(additional_parts):
        if not additional_item:
            continue
        additional_content = extract_additional_content(
            repository_path, chapter, additional_item
        )
        if not additional_content:
            continue
        additional_contents.append(additional_content)
    return additional_contents


def include_additional_contents(repository_path, chapter, html_content):
    # Useful to be able to match the `contenuadd_pattern`, otherwise
    # escaped special chars will not match:
    # "EnvironnementNumérique" vs. "EnvironnementNum&#xE9;rique".
    html_content = html.unescape(html_content)
    chapter.additional_contents = (
        chapter.additional_contents if hasattr(chapter, "additional_contents") else []
    )
    for additional_content in extract_additional_contents(repository_path, chapter):
        contenuadd_pattern = f"!contenuadd(./{additional_content.id})"
        html_content = html_content.replace(
            f"<p>{contenuadd_pattern}</p>",
            f"<!-- From: {contenuadd_pattern} -->\n{additional_content.html}",
        )
        chapter.additional_contents.append(additional_content)
    return html_content



import contextlib
import os
import shutil
import socket
import zipfile
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer, test
from pathlib import Path
from typing import Optional

import httpx
import yaml
from minicli import cli, run

from . import ROOT_DIR, VERSION
from .corpus import convert_stylo_articles, generate_book_metadata
from .generator import (
    generate_chapters,
    generate_homepage,
    generate_markdown,
    generate_pdf,
)
from .indexes import generate_indexes
from .models import configure_book
from .search import generate_search_page
from .statics import bundle_statics, sync_media, sync_statics


@cli
def version():
    """Return the current version of pressoir."""
    print(f"Pressoir version: {VERSION}")


@cli
@cli("collection", choices=["pum", "sp", "blank"])
def init(repository_path: Path = Path(), collection: str = "blank"):
    """Initialize a new book to `repository_path` or current directory.

    :repository_path: Absolute or relative path to book’s sources (default: current).
    :collection: Name of the collection (default: blank).
    """
    print(
        f"Initializing a new book: `{repository_path}` for `{collection}` collection."
    )

    if not (repository_path / "presssoir").exists():
        shutil.copytree(
            ROOT_DIR / "init" / collection / "pressoir",
            repository_path / "pressoir",
            dirs_exist_ok=True,
        )

    if "textes" not in os.listdir(repository_path):
        shutil.copytree(
            ROOT_DIR / "init" / collection / "textes",
            repository_path / "textes",
            dirs_exist_ok=True,
        )

    if "doc" not in os.listdir(repository_path) and "doc" in os.listdir(
        ROOT_DIR / "init" / collection
    ):
        shutil.copytree(
            ROOT_DIR / "init" / collection / "doc",
            repository_path / "doc",
            dirs_exist_ok=True,
        )


@cli
def docs(target_path: Optional[Path] = None):
    """Generate documentation with pressoir itself. #SoMeta"""
    if target_path is None:
        target_path = Path(os.getenv("PWD", "")) / "public"
    else:
        target_path = Path(target_path)
    print(f"Generating documentation in `{target_path.resolve()}`.")
    build(ROOT_DIR / "docs", target_path=target_path)  # , verbose=True)
    print("Don’t forget to generate the associated PDF file :)")
    print("pressoir export --repository-path=pressoir/docs")


@cli
def build(
    repository_path: Path = Path(),
    csl_path: Optional[Path] = None,
    target_path: Optional[Path] = None,
    chapter: str = "",
    keep_statics: bool = False,
    verbose: bool = False,
):
    """Build a book from `repository_path` or current directory.

    :repository_path: Absolute or relative path to book’s sources (default: current).
    :csl_path: Path to .csl file (default: Pandoc’s default).
    :target_path: Where the book will be built (default: `repository_path`/public).
    :chapter: Specify a given chapter id (e.g. `chapter1`).
    :keep_statics: Do not override the statics with regular ones (default: False).
    :verbose: Display more informations during the build (default: False).
    """
    if target_path is None:
        target_path = repository_path / "public"
    else:
        target_path = Path(target_path)
    target_path.mkdir(parents=True, exist_ok=True)
    print(
        f"Building a book from {repository_path.resolve()} to {target_path.resolve()}."
    )
    book = configure_book(repository_path / "textes" / "garde" / "livre.yaml")
    if not keep_statics:
        sync_media(repository_path, target_path, book)
        sync_statics(repository_path, target_path)
    css_filename, js_filename = bundle_statics(repository_path, target_path)
    if verbose:
        import pprint

        pprint.pprint(book)

    meta = {"css_filename": css_filename, "js_filename": js_filename}
    generate_homepage(repository_path, target_path, book, meta)
    chapters = generate_chapters(
        repository_path, csl_path, target_path, book, meta, chapter
    )
    generate_indexes(repository_path, target_path, book)
    generate_search_page(repository_path, target_path, book, chapters)


@cli
def export(
    repository_path: Path = Path(),
    template_path: Optional[Path] = None,
    csl_path: Optional[Path] = None,
    target_path: Optional[Path] = None,
    verbose: bool = False,
):
    """Generate a single md+tex+pdf file from `repository_path` or current directory.

    :repository_path: Path to book’s sources (default: current).
    :template_path: Path to .tex template (default: Pandoc’s default).
    :csl_path: Path to .csl file (default: Pandoc’s default).
    :target_path: Where the book will be built (default: `repository_path`/public).
    :verbose: Display a lot of informations, useful for debugging.
    """
    if target_path is None:
        target_path = repository_path / "public"
    else:
        target_path = Path(target_path)
    target_path.mkdir(parents=True, exist_ok=True)
    print(
        f"Generating file from {repository_path.resolve()} to {target_path.resolve()}."
    )
    book = configure_book(repository_path / "textes" / "garde" / "livre.yaml")
    if verbose:
        import pprint

        pprint.pprint(book)
    generate_markdown(repository_path, target_path, book)
    generate_pdf(repository_path, template_path, csl_path, target_path, book)


@cli
def serve(repository_path: Path = Path(), port: int = 8000):
    """Serve an HTML book from `repository_path`/public or current directory/public.

    :repository_path: Absolute or relative path to book’s sources (default: current).
    :port: Port to serve the book from (default=8000)
    """
    print(
        f"Serving HTML book from `{repository_path}/public` to http://127.0.0.1:{port}"
    )

    # From https://github.com/python/cpython/blob/main/Lib/http/server.py#L1307-L1326
    class DirectoryServer(ThreadingHTTPServer):
        def server_bind(self):
            # suppress exception when protocol is IPv4
            with contextlib.suppress(Exception):
                self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            return super().server_bind()

        def finish_request(self, request, client_address):
            self.RequestHandlerClass(
                request, client_address, self, directory=str(repository_path / "public")
            )

    test(HandlerClass=SimpleHTTPRequestHandler, ServerClass=DirectoryServer, port=port)


@cli
def stylo(
    stylo_id: str,
    stylo_instance: str = "stylo.huma-num.fr",
    stylo_export: str = "https://export.stylo.huma-num.fr",
    from_scratch: bool = False,
    keep_metadata: bool = False,
):
    """Initialize a new book to current directory from Stylo.

    :stylo_id: Corpus id from Stylo.
    :stylo_instance: Instance of Stylo (default: stylo.huma-num.fr).
    :stylo_export: Stylo export URL (default: https://export.stylo.huma-num.fr).
    :from_scratch: Do not ask to override local files (default: False).
    :keep_metadata: Do not override the `livre.yaml` metadata file (default: False).
    """
    print(
        f"Initializing a new corpus: `{stylo_id}` from `{stylo_instance}` "
        f"through export service `{stylo_export}`."
    )

    if not (Path() / "presssoir").exists():
        print("Copying Pressoir-related files from `blank` init.")
        shutil.copytree(
            ROOT_DIR / "init" / "blank" / "pressoir",
            Path() / "pressoir",
            dirs_exist_ok=True,
        )

    textes_path = Path() / "textes"
    if not textes_path.exists():
        Path.mkdir(textes_path)
    else:
        if (
            from_scratch
            or input("Écraser les textes locaux avec ceux de Stylo ? (O / N) ") == "O"
        ):
            print("Erasing existing local textes…")
            shutil.rmtree(textes_path)
            Path.mkdir(textes_path)
        else:
            print("Keeping existing local textes. Nothing updated from Stylo.")
            return

    url = (
        f"{stylo_export}/generique/corpus/export/"
        f"{stylo_instance}/{stylo_id}/Extract-corpus/"
        "?with_toc=0&with_ascii=0&with_link_citations=0&with_nocite=0"
        "&version=&bibliography_style=chicagomodified&formats=originals"
    )
    zip_path = Path() / f"export-{stylo_id}.zip"
    print(f"Downloading data from {url} to {zip_path}")
    with Path.open(zip_path, "wb") as fd:
        with httpx.stream("GET", url, timeout=None) as r:
            for data in r.iter_bytes():
                fd.write(data)

    try:
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(textes_path)
        print(f"Data downloaded and extracted to {textes_path}")
    except zipfile.BadZipFile:
        print(f"Unable to find corpus with id {stylo_id}!")
        return

    convert_stylo_articles(textes_path, pressoir_dirs=["garde", "media", "corpus"])

    Path.mkdir(textes_path / "garde", exist_ok=True)
    shutil.copy(
        ROOT_DIR / "init" / "blank" / "textes" / "garde" / "homepage.md",
        textes_path / "garde",
    )

    if keep_metadata:
        return

    print("Generating metadata for book")
    corpus_data = next(yaml.safe_load_all((textes_path / "corpus.yaml").read_text()))
    print(f"From corpus data: {corpus_data}")
    book_metadata = generate_book_metadata(textes_path, corpus_data)
    print(f"To pressoir data: {book_metadata}")
    book_yaml = yaml.dump(book_metadata)
    with Path.open(textes_path / "garde" / "livre.yaml", "w") as metadata_file:
        metadata_file.write(book_yaml)


def main():
    run()



import json
import re

from jinja2 import Environment as Env
from jinja2 import FileSystemLoader
from jinja2.filters import do_striptags

from . import ROOT_DIR

environment = Env(loader=FileSystemLoader(str(ROOT_DIR / "templates")))


RE_CONTENUS_ADD = re.compile(r"!contenuadd(.*)")


def generate_search(search_index, book):
    template = environment.get_template("recherche.html")
    return template.render(search_index=search_index, book=book)


def write_search(output, content):
    index_html = output.read_text()
    index_html = index_html.replace("%SEARCH%", content)
    output.write_text(index_html)


def clean_html(html_content):
    html_content = re.sub(RE_CONTENUS_ADD, "", html_content)
    return (
        do_striptags(html_content)
        .replace("'", " ")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
    )


def generate_search_page(repository_path, target_path, book, chapters):
    """Generate the search page and related index."""
    print(f"Initializing search for {repository_path}")
    search_index = []
    for chapter in chapters:
        if chapter.id not in ["index-np", "recherche"]:
            search_index.append(
                {
                    "book_title": book.title,
                    "book_url": book.url,
                    "book_coverurl": book.coverurl,
                    "title": chapter.title,
                    "url": chapter.url_relative,
                    "type": "chapitre",
                    "content": clean_html(chapter.html),
                }
            )
            for additional_content in chapter.additional_contents:
                search_index.append(
                    {
                        "book_title": book.title,
                        "book_url": book.url,
                        "book_coverurl": book.coverurl,
                        "title": additional_content.title,
                        "url": f"{chapter.url_relative}#{additional_content.id}",
                        "type": "contenu additionnel",
                        "content": clean_html(additional_content.html),
                    }
                )
    search_json = json.dumps(search_index, indent=2)
    (target_path / "recherche.json").write_text(search_json)
    search_content = generate_search(search_json, book)
    search_file = target_path / "recherche.html"
    if search_file.exists():
        write_search(search_file, search_content)



import shutil

from slugify import slugify

from . import ROOT_DIR
from .utils import each_file_from, generate_md5

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib


def sync_media(repository_path, target_path, book):
    if not (repository_path / "pressoir").exists():
        print(f"The `pressoir` folder is missing from `{repository_path}`.")
        exit(1)

    media_path = repository_path / "textes" / "media"
    if media_path.exists():
        shutil.copytree(media_path, target_path / "media", dirs_exist_ok=True)
    book_settings = tomllib.loads(
        (repository_path / "pressoir" / "book.toml").read_text()
    )
    # Copy the PDF file if defined in `pressoir/book.toml` as `media.pdf_path`
    # to the `pdfurl` target in `livre.yaml` if defined too.
    if media := book_settings.get("media"):
        if pdf_path := media.get("pdf_path"):
            if pdf_url := book.pdfurl.lstrip("./"):
                shutil.copy(repository_path / pdf_path, target_path / pdf_url)


def sync_statics(repository_path, target_path):
    if not (repository_path / "pressoir").exists():
        print(f"The `pressoir` folder is missing from `{repository_path}`.")
        exit(1)
    shutil.copytree(
        repository_path / "pressoir" / "static" / "fonts",
        target_path / "static" / "fonts",
        dirs_exist_ok=True,
    )
    shutil.copytree(
        ROOT_DIR / "static" / "svg",
        target_path / "static" / "svg",
        dirs_exist_ok=True,
    )
    shutil.copytree(
        repository_path / "pressoir" / "static" / "svg",
        target_path / "static" / "svg",
        dirs_exist_ok=True,
    )
    shutil.copytree(
        ROOT_DIR / "static" / "img",
        target_path / "static" / "img",
        dirs_exist_ok=True,
    )
    img_path = repository_path / "pressoir" / "static" / "img"
    if img_path.exists():
        shutil.copytree(img_path, target_path / "static" / "img", dirs_exist_ok=True)

    book_settings = tomllib.loads(
        (repository_path / "pressoir" / "book.toml").read_text()
    )
    if theme := book_settings.get("theme"):
        for svg_file in each_file_from(target_path / "static" / "svg", pattern="*.svg"):
            replacements = theme.get(slugify(svg_file.name))
            if replacements:
                svg_content = svg_file.read_text()
                svg_content = svg_content.replace(replacements[0], replacements[1])
                svg_file.write_text(svg_content)


def bundle_statics(repository_path, target_path):
    pressoir_settings = tomllib.loads((ROOT_DIR / "project.toml").read_text())
    book_settings = tomllib.loads(
        (repository_path / "pressoir" / "book.toml").read_text()
    )
    theme = book_settings.get("theme")
    indexes = book_settings.get("indexes")

    css_parts = []
    for css_file in pressoir_settings["statics"]["css_files"]:
        css_parts.append(f"/* {css_file} */")
        css_parts.append((ROOT_DIR / css_file).read_text())
    for css_file in book_settings["statics"]["css_files"]:
        css_parts.append(f"/* {css_file} */")
        css_content = (repository_path / "pressoir" / css_file).read_text()
        # Dynamic CSS variables replacements from `book.toml[theme]`.
        if theme and css_file.endswith("vars.css"):
            for key, [initial, replacement] in theme.items():
                source = f"--{key}: {initial};"
                target = f"--{key}: {replacement};"
                css_content = css_content.replace(source, target)
        # Dynamic CSS addition for indexes from `book.toml[indexes]`.
        if indexes and css_file.endswith("custom.css"):
            for index, image in zip(indexes["ids"], indexes["images"]):
                css_content = f"""
                    .termeenrichi.{index} {{
                        background-image: url("{image}");
                    }}
                    {css_content}
                """
        css_parts.append(css_content)

    js_parts = []
    for js_file in pressoir_settings["statics"]["js_files"]:
        js_parts.append(f"/* {js_file} */")
        js_parts.append((ROOT_DIR / js_file).read_text())
    for js_file in book_settings["statics"].get("js_files", []):
        js_parts.append(f"/* {js_file} */")
        js_parts.append((repository_path / "pressoir" / js_file).read_text())

    css_content = "\n".join(css_parts)
    css_filename = f"bundle.{generate_md5(css_content)}.css"
    (target_path / "static" / css_filename).write_text(css_content)

    js_content = "\n".join(js_parts)
    js_filename = f"bundle.{generate_md5(js_content)}.js"
    (target_path / "static" / js_filename).write_text(js_content)

    return css_filename, js_filename



import hashlib
import re
from pathlib import Path

from . import ROOT_DIR

html_comments_re = re.compile(r"<!--.*?-->")
html_tags_re = re.compile(r"<[^>]*>")


def strip_html_comments(value):
    if not value:
        return ""
    return html_comments_re.sub("", value)


def strip_html_tags(value):
    if not value:
        return ""
    return html_tags_re.sub("", value)


def get_template_path(repository_path, name):
    """Try to find it within the book folder, fallback on pressoir one."""
    template_path = repository_path / "templates" / name
    if not template_path.exists():
        template_path = ROOT_DIR / "templates" / name
    return template_path


def each_file_from(source_dir, pattern="*.html", exclude=None):
    """Walk across the `source_dir` and return the html file paths."""
    for path in _each_path_from(source_dir, pattern=pattern, exclude=exclude):
        if path.is_file():
            yield path


def each_folder_from(source_dir, exclude=None):
    """Walk across the `source_dir` and return the folder paths."""
    for path in _each_path_from(source_dir, exclude=exclude):
        if path.is_dir():
            yield path


def _each_path_from(source_dir, pattern="*", exclude=None):
    for path in sorted(Path(source_dir).glob(pattern)):
        if exclude is not None and path.name in exclude:
            continue
        yield path


def generate_md5(content):
    return hashlib.md5(content.encode()).hexdigest()


def neighborhood(iterable, first=None, last=None):
    """
    Yield the (index, previous, current, next) items given an iterable.

    You can specify a `first` and/or `last` item for bounds.
    """
    index = 1
    iterator = iter(iterable)
    previous = first
    current = next(iterator)  # Throws StopIteration if empty.
    for next_ in iterator:
        yield (index, previous, current, next_)
        previous = current
        index += 1
        current = next_
    yield (index, previous, current, last)



from selectolax.parser import HTMLParser


def convert_sidenotes(html_content):
    """Converts sidenotes for TufteCSS."""
    parsed_html = HTMLParser(html_content)
    sidenotes = parsed_html.css(".footnote-ref")
    for count, sidenote in enumerate(sidenotes, start=1):
        # Get the first parent section, Pandoc generates invalid HTML
        # with duplicated ids for sidenotes when generated for main
        # content and additional contents. We try to be more selective
        # and generate our own custom sidenote ref.
        custom_sidenote_id = f"sidenote-{count}"
        custom_sidenote_ref = f"sidenote-ref-{count}"
        sidenote_id = sidenote.attributes["id"]

        parent_section = sidenote.parent
        while parent_section.css_first(sidenote.attributes["href"]) is None:
            parent_section = parent_section.parent

        sidenote_ref = parent_section.css_first(sidenote.attributes["href"])
        sidenote_ref.unwrap_tags(["p"])
        sidenote_content = sidenote_ref.html
        sidenote_content = sidenote_content.replace("<li ", '<span class="sidenote"')
        sidenote_content = sidenote_content.replace("</li>", "</span>")
        parsed_sidenote_content = HTMLParser(sidenote_content)
        sidenote_ref_id = parsed_sidenote_content.body.child.attributes["id"]
        sidenote_content = sidenote_content.replace(
            f'id="{sidenote_ref_id}"', f'id="{custom_sidenote_id}"'
        )
        sidenote_content = sidenote_content.replace(
            f'href="#{sidenote_id}"', f'href="#{custom_sidenote_ref}"'
        )
        # replace_with from selectolax only allows a Node,
        # otherwise it will escape the tags! Hence the extra-parsing.
        tufte_sidenote = HTMLParser(
            (
                # Do not add more spaces between HTML elements otherwise
                # it leads to orphans in the body of the text.
                f"<span>"
                f'<label for="{custom_sidenote_ref}" class="margin-toggle '
                'sidenote-number"></label>'
                f'<input type="checkbox" id="{custom_sidenote_ref}" '
                'class="margin-toggle">'
                f"{sidenote_content}"
                f"</span>"
            )
        )
        tufte_sidenote_span = tufte_sidenote.css_first("span")
        sidenote.replace_with(tufte_sidenote_span)
    for footnote in parsed_html.css(".footnotes"):
        footnote.remove()
    return parsed_html.html


def rewrite_global_sidenotes(html_content):
    # We try to avoid ids matching the ones from additional contents.
    html_content = (
        html_content.replace('id="fn', 'id="global-fn')
        .replace('href="#fn', 'href="#global-fn')
        .replace('id="fnref', 'id="global-fnref')
        .replace('href="#fnref', 'href="#global-fnref')
    )
    return html_content



import re
import tempfile

import pypandoc
import yaml
from dataclass_wizard.dumpers import asdict
from slugify import slugify

from .additionals import include_additional_contents
from .models import Part
from .sidenotes import convert_sidenotes, rewrite_global_sidenotes
from .utils import get_template_path, neighborhood

try:
    print(f"Pandoc version: {pypandoc.get_pandoc_version()}")
except OSError:
    pypandoc.download_pandoc()
    print(f"Pandoc version: {pypandoc.get_pandoc_version()}")

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

RE_CONTENUS_ADD = re.compile(r"\n\n!contenuadd(.*)")


def generate_chapters(repository_path, csl_path, target_path, book, meta, chapter_id):
    print("  Generating chapters:")
    book_settings = tomllib.loads(
        (repository_path / "pressoir" / "book.toml").read_text()
    )

    indexes_ids = book_settings.get("indexes", {}).get("ids")
    meta["indexes"] = indexes_ids
    glossaire = book_settings.get("glossaire", {})
    meta["glossaire"] = {
        "button": {
            "label": glossaire.get("button-label", "Voir dans le glossaire"),
            "title": glossaire.get(
                "button-title", "Consulter la référence dans le glossaire de l’ouvrage"
            ),
        }
    }
    meta["nb_of_chapters"] = len(book.chapters)
    indexed_chapters = []

    for index, previous_chapter, current_chapter, next_chapter in neighborhood(
        book.chapters
    ):
        meta["index"] = index
        if not chapter_id or chapter_id == current_chapter.id:
            chapter = generate_chapter(
                repository_path,
                csl_path,
                target_path,
                book,
                current_chapter,
                previous_chapter,
                next_chapter,
                meta,
            )
            indexed_chapters.append(chapter)
    return indexed_chapters


def generate_chapter(
    repository_path,
    csl_path,
    target_path,
    book,
    current_chapter,
    previous_chapter,
    next_chapter,
    meta,
):
    print(
        f"    {current_chapter.id}: {current_chapter.title_h} "
        + f"({meta['index']}/{meta['nb_of_chapters']})"
    )
    dict_content = {
        "book": asdict(book),
        "chapter": asdict(current_chapter),
        "prev": (previous_chapter and asdict(previous_chapter)) or "",
        "next": (next_chapter and asdict(next_chapter)) or "",
        "meta": meta,
    }
    yaml_content = yaml.dump(dict_content)
    yaml_content += "nocite: '[@*]'\n"
    with tempfile.NamedTemporaryFile() as metadata_file:
        metadata_file.write(yaml_content.encode("utf-8"))
        metadata_file.read()  # Required to be readable by Pandoc.
        html_content = generate_html_content(
            repository_path, csl_path, book, current_chapter, metadata_file
        )
        # This is fragile and depends on templates.
        _, main = html_content.split("<main>", 1)
        content, _ = main.split("</main>", 1)
        current_chapter.html = content
    html_content = rewrite_global_sidenotes(html_content)
    html_content = include_additional_contents(
        repository_path, current_chapter, html_content
    )
    html_content = convert_sidenotes(html_content)
    (target_path / f"{current_chapter.id}.html").write_text(html_content)
    return current_chapter


def generate_html_content(
    repository_path, csl_path, book, current_chapter, metadata_file
):
    textes_path = repository_path / "textes"
    chapter_id = current_chapter.id
    md_content = (textes_path / chapter_id / f"{chapter_id}.md").read_text()
    bib_file = textes_path / chapter_id / f"{chapter_id}.bib"
    zotero_collection = current_chapter.zoteroCollection
    zotero_description = book.zoteroDescription
    extra_refs = ""
    if zotero_collection:
        extra_refs = f"""
<a href="{zotero_collection}" title="Collection Zotero du chapitre" class="zotero">
    <strong>{zotero_description}</strong>
</a>
        """
    md_content = md_content.replace(
        "## Références",
        f"""
<section>
<details class="references" open>
<summary id="references">Références</summary>

{extra_refs}

:::{{#refs}}
:::

</details>
</section>""",
    )

    template_path = get_template_path(repository_path, "chapitre.html")
    lang = current_chapter.lang or book.lang or "fr"
    extra_args = [
        "--ascii",
        "--citeproc",
        f"--bibliography={bib_file}",
        f"--template={template_path}",
        f"--metadata=lang:{lang}",
        f"--metadata=title:{current_chapter.title_f}",
        f"--metadata-file={metadata_file.name}",
    ]
    if csl_path is not None:
        extra_args.append(f"--csl={csl_path}")

    html_content = pypandoc.convert_text(
        md_content,
        "html",
        format="markdown+auto_identifiers+ascii_identifiers",
        extra_args=extra_args,
    )

    return html_content


def generate_homepage(repository_path, target_path, book, meta):
    print(f"Rendering {target_path}:")
    print("  Generating homepage")

    source = repository_path / "textes" / "garde" / "homepage.md"
    if not source.exists():
        print("    Homepage not generated (missing homepage.md file)")
        return
    dict_content = {
        "book": asdict(book),
        "meta": meta,
    }
    yaml_content = yaml.dump(dict_content)
    yaml_content += "nocite: '[@*]'\n"
    with tempfile.NamedTemporaryFile() as metadata_file:
        metadata_file.write(yaml_content.encode("utf-8"))
        metadata_file.read()  # Required to be readable by Pandoc.
        md_content = (repository_path / "textes" / "garde" / "homepage.md").read_text()
        template_path = get_template_path(repository_path, "homepage.html")
        extra_args = [
            "--ascii",
            "--citeproc",
            f"--template={template_path}",
            f"--metadata=title:{book.title_f}",
            f"--metadata=first_chapter_id:{dict_content['book']['chapters'][0]['id']}",
            f"--metadata-file={metadata_file.name}",
        ]

        # By default, pressoir assumes there are multipe authors.
        # Check whether there is more than one author, and pass an extra
        # variable as a cli argument to pandoc. A variable with no specified
        # value is automatically assigned to true. This exists to work around a
        # limitation in pandoc templates, where you cannot make complex
        # conditional statements, but only check for variable truthfullness.
        if len(dict_content['book']['authors']) == 1:
            extra_args.append(f"--variable=singleauthor")

        html_content = pypandoc.convert_text(
            md_content, "html", format="md", extra_args=extra_args
        )

    (target_path / "index.html").write_text(html_content)


def prepare_chapter_markdown(
    repository_path, filenames_to_slugs, chapter_or_part, within_part=False
):
    textes_path = repository_path / "textes"
    chapter_id = chapter_or_part.id

    if chapter_id in ["bibliography", "index-np", "recherche"]:
        return None, None

    # If the chapter is within a part we want to start at the second level
    # and pass on to subsequent titles the new hierarchy.
    chapter_title_level = "#"
    if within_part:
        chapter_title_level = "##"

    md_content = (textes_path / chapter_id / f"{chapter_id}.md").read_text()

    # Fix internal references.
    for filename, slug in filenames_to_slugs.items():
        # First we consider references with particular anchor.
        md_content = md_content.replace(f"]({filename}#", "](#")
        # Then we deal with link to whole pages.
        md_content = md_content.replace(f"]({filename}", f"](#{slug}")

    # Remove contenus additionnels.
    md_content = re.sub(RE_CONTENUS_ADD, "", md_content)
    # Remove everything below references completely
    md_content = md_content.split("\n## Références\n", 1)[0]
    # Avoid numerotation of titles + dynamic titles hierarchy.
    md_content = re.sub(r"## (.*)", rf"{chapter_title_level}# \1 {{-}}", md_content)

    bib_path = textes_path / chapter_id / f"{chapter_id}.bib"
    bib_content = bib_path.read_text()

    # Re-number footnotes to avoid duplicates across chapters.
    md_content = re.sub(r"\[\^(\d+)\]", rf"[^{chapter_id}-\g<1>]", md_content)
    # Add the (unumbered) title per chapter.
    md_content = f"""
{chapter_title_level} {chapter_or_part.title_h} {{-}}

{md_content}
    """
    return md_content, bib_content


def generate_markdown(repository_path, target_path, book):
    print(f"Exporting {target_path} in Markdown:")
    md_contents = ""
    bib_contents = ""

    filenames_to_slugs = {
        f"{chapter.id}.html": slugify(chapter.title_h) for chapter in book.chapters
    }

    for chapter_or_part in book.chapters_and_parts:
        if isinstance(chapter_or_part, Part):
            md_contents += f"\n# {chapter_or_part.title} {{-}}\n\n"

            for chapter in chapter_or_part.chapters:
                if not chapter.pdf:
                    print(f"  Chapter excluded (pdf=false): {chapter.title}")
                    continue

                md_content, bib_content = prepare_chapter_markdown(
                    repository_path, filenames_to_slugs, chapter, within_part=True
                )
                if md_content:
                    md_contents += md_content
                if bib_content:
                    bib_contents += bib_content
        else:  # This is a Chapter.
            if not chapter_or_part.pdf:
                print(f"  Chapter excluded (pdf=false): {chapter_or_part.title}")
                continue

            md_content, bib_content = prepare_chapter_markdown(
                repository_path, filenames_to_slugs, chapter_or_part
            )
            if md_content:
                md_contents += md_content
            if bib_content:
                bib_contents += bib_content

    (target_path / "book.md").write_text(md_contents)
    print(f"  Markdown written in: {target_path / 'book.md'}")
    (target_path / "book.bib").write_text(bib_contents)
    print(f"  Bibliography written in: {target_path / 'book.bib'}")


def generate_pdf(repository_path, template_path, csl_path, target_path, book):
    print(f"Exporting {target_path} in PDF:")
    md_content = (target_path / "book.md").read_text()
    bib_path = target_path / "book.bib"

    extra_args = [
        f"--metadata-file={repository_path / 'textes' / 'garde' / 'livre.yaml'}",
        f"--resource-path={repository_path}:{repository_path / 'textes'}",
        f"--bibliography={bib_path}",
        "--citeproc",
        "--standalone",
    ]
    if book.has_parts:
        extra_args.append("--top-level-division=part")
    else:
        extra_args.append("--top-level-division=chapter")

    if template_path is not None:
        extra_args.append(f"--template={template_path}")
    if csl_path is not None:
        extra_args.append(f"--csl={csl_path}")
    latex_content = pypandoc.convert_text(
        md_content,
        "latex",
        format="markdown+auto_identifiers+ascii_identifiers",
        extra_args=extra_args,
    )
    (target_path / "book.tex").write_text(latex_content)
    print(f"  TEX written in: {target_path / 'book.tex'}")
    extra_args = [
        f"--metadata-file={repository_path / 'textes' / 'garde' / 'livre.yaml'}",
        f"--resource-path={repository_path}:{repository_path / 'textes'}",
        f"--bibliography={bib_path}",
        "--citeproc",
        "--top-level-division=chapter",
        "--pdf-engine=lualatex",
    ]
    if book.has_parts:
        extra_args.append("--top-level-division=part")
    else:
        extra_args.append("--top-level-division=chapter")

    if template_path is not None:
        extra_args.append(f"--template={template_path}")
    pypandoc.convert_text(
        latex_content,
        "pdf",
        format="latex",
        extra_args=extra_args,
        outputfile=str(target_path / "book.pdf"),
    )
    print(f"  PDF written in: {target_path / 'book.pdf'}")



import re
import shutil

import yaml
from slugify import slugify

from .utils import each_file_from, each_folder_from


def generate_book_metadata(textes_path, corpus_data):
    corpus_metadata = corpus_data.get("metadata") or {}
    articles_titles = [
        slugify(article["article"]["title"]) for article in corpus_data.get("articles")
    ]
    articles_authors = []
    for article in corpus_data.get("articles"):
        for yaml_file in each_file_from(
            textes_path / article["article"]["title"], pattern="*.yaml"
        ):
            article_data = next(yaml.safe_load_all(yaml_file.read_text()))
            articles_authors += article_data.get("authors", [])

    book_metadata = {
        "title": corpus_metadata.get("issue", {}).get("title", "Titre à définir"),
        "version": corpus_metadata.get("@version"),
        "lang": "fr",
        "date": "2025-01-28",
        "rights": (
            "Creative Commons Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)"
        ),
        "url": "",
        "collective": False,
        "coverurl": "",
        "abstract_fr": "Résumé",
        "abstract_en": "Abstract",
        "keyword_fr": "",
        "keyword_en": "",
        "isbnprint": "",
        "isbnepub": "",
        "isbnpdf": "",
        "isbnnum": "",
        "publisher": "Pressoir éditions",
        "place": "Montréal",
        "authors": articles_authors,
        "toc": [{"id": article_title} for article_title in articles_titles],
    }
    return book_metadata


def rewrite_image_paths(md_file):
    print(f"Rewriting image paths from `images/` to `media/` in {md_file}")
    md_file.write_text(md_file.read_text().replace("images/", "media/"))


RE_ADD_CONTENTS = re.compile(
    r"(?s):{3}\s*{\s?\.contenuadd\s*#(?P<id>.*?)\s?}\n(?P<content>.*?)\n:{3}"
)


def extract_additional_contents(md_file):
    md_content = md_file.read_text()
    matcher = ":::"

    if matcher in md_content:
        additional_contents = ""
        additional_references = []
        for match in RE_ADD_CONTENTS.finditer(md_content):
            id_ = match.group("id").strip()
            content = match.group("content").strip()
            additional_contents += f"## {id_}\n\n{content}\n\n"
            additional_references.append([id_, match.span()])
        (md_file.parent / "additionnels.md").write_text(additional_contents)
        # Reverse is important to keep positions while iterating.
        for id_, span in reversed(additional_references):
            md_content = (
                md_content[: span[0]] + f"!contenuadd(./{id_})" + md_content[span[1] :]
            )
        md_file.write_text(md_content)


def convert_stylo_articles(textes_path, pressoir_dirs):
    for folder in each_folder_from(textes_path):
        if folder.name in pressoir_dirs:
            continue
        for subfolder in each_folder_from(folder):
            if subfolder.name == "images":
                print("Copying images from `images/` to `media/`")
                shutil.copytree(subfolder, textes_path / "media", dirs_exist_ok=True)
                for md_file in each_file_from(folder, pattern="*.md"):
                    rewrite_image_paths(md_file)
        for md_file in each_file_from(folder, pattern="*.md"):
            extract_additional_contents(md_file)

        target = folder.parent / "-".join(folder.name.split("-")[:-1])
        print(f"Renaming folder {folder} to {target}")
        folder.rename(target)
